from __future__ import annotations

import io
from typing import IO, TYPE_CHECKING

from .binary import encode_bin, decode_bin

if TYPE_CHECKING:
    from typing_extensions import Buffer  # 3.12+
    from typing_extensions import Self  # 3.11+


class BitStream(io.RawIOBase, IO[bytes]):

    __slots__ = ("substream",)

    def __init__(self, substream: IO[bytes]) -> None:
        self.substream = substream

    def __enter__(self) -> Self:
        return self


class BitStreamReader(BitStream):

    __slots__ = ("buffer", "total_size")

    def __init__(self, substream: IO[bytes]) -> None:
        super().__init__(substream)
        self.total_size = 0
        self.buffer = b""

    def close(self) -> None:
        if self.total_size % 8 != 0:
            raise ValueError("total size of read data must be a multiple of 8",
                self.total_size)

    def tell(self) -> int:
        return self.substream.tell()

    def seek(self, pos: int, whence: int = 0) -> int:
        self.buffer = b""
        self.total_size = 0
        self.substream.seek(pos, whence)
        return 0

    def read(self, count: int = -1) -> bytes:
        if count < 0:
            raise ValueError("count cannot be negative")

        l = len(self.buffer)
        if count == 0:
            data = b""
        elif count <= l:
            data = self.buffer[:count]
            self.buffer = self.buffer[count:]
        else:
            data = self.buffer
            count -= l
            bytes = count // 8
            if count & 7:
                bytes += 1
            buf = encode_bin(self.substream.read(bytes))
            data += buf[:count]
            self.buffer = buf[count:]
        self.total_size += len(data)
        return data


class BitStreamWriter(BitStream):

    __slots__ = ("buffer", "pos")

    def __init__(self, substream: IO[bytes]) -> None:
        super().__init__(substream)
        self.buffer: list[bytes] = []
        self.pos = 0

    def close(self) -> None:
        self.flush()

    def flush(self) -> None:
        bytes = decode_bin(b"".join(self.buffer))
        self.substream.write(bytes)
        self.buffer = []
        self.pos = 0

    def tell(self) -> int:
        return self.substream.tell() + self.pos // 8

    def seek(self, pos: int, whence: int = 0) -> int:
        self.flush()
        return self.substream.seek(pos, whence)

    def write(self, data: Buffer) -> int:
        if not data:
            return 0
        if type(data) is not bytes:
            raise TypeError("data must be a bytes, not %r" % (type(data),))
        self.buffer.append(data)
        return len(data)



#-------------------------------------------------------------------------------
# elftools: dwarf/locationlists.py
#
# DWARF location lists section decoding (.debug_loc)
#
# Eli Bendersky (eliben@gmail.com)
# This code is in the public domain
#-------------------------------------------------------------------------------
from __future__ import annotations

import os
from typing import IO, TYPE_CHECKING, NamedTuple

from ..common.exceptions import DWARFError
from ..common.utils import struct_parse
from .dwarf_util import _iter_CUs_in_section

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator, Mapping

    from ..construct.lib.container import Container
    from .compileunit import CompileUnit
    from .die import DIE, AttributeValue
    from .dwarfinfo import DWARFInfo
    from .structs import DWARFStructs
    from .typeunit import TypeUnit


class LocationExpr(NamedTuple):
    loc_expr: list[int]


class LocationEntry(NamedTuple):
    entry_offset: int
    entry_length: int
    begin_offset: int
    end_offset: int
    loc_expr: list[int]
    is_absolute: bool


class BaseAddressEntry(NamedTuple):
    entry_offset: int
    entry_length: int
    base_address: int


class LocationViewPair(NamedTuple):
    entry_offset: int
    begin: int
    end: int


if TYPE_CHECKING:
    _Location = LocationExpr | LocationEntry | BaseAddressEntry | LocationViewPair


def _translate_startx_length(e: Container, cu: CompileUnit | TypeUnit) -> LocationEntry:
    start_offset: int = cu.dwarfinfo.get_addr(cu, e.start_index)
    return LocationEntry(e.entry_offset, e.entry_length, start_offset, start_offset + e.length, e.loc_expr, True)

# Maps parsed entries to the tuples above; LocationViewPair is mapped elsewhere
entry_translate: dict[str, Callable[[Container, CompileUnit | TypeUnit], _Location]] = {
    'DW_LLE_base_address'    : lambda e, cu: BaseAddressEntry(e.entry_offset, e.entry_length, e.address),
    'DW_LLE_offset_pair'     : lambda e, cu: LocationEntry(e.entry_offset, e.entry_length, e.start_offset, e.end_offset, e.loc_expr, False),
    'DW_LLE_start_length'    : lambda e, cu: LocationEntry(e.entry_offset, e.entry_length, e.start_address, e.start_address + e.length, e.loc_expr, True),
    'DW_LLE_start_end'       : lambda e, cu: LocationEntry(e.entry_offset, e.entry_length, e.start_address, e.end_address, e.loc_expr, True),
    'DW_LLE_default_location': lambda e, cu: LocationEntry(e.entry_offset, e.entry_length, -1, -1, e.loc_expr, True),
    'DW_LLE_base_addressx'   : lambda e, cu: BaseAddressEntry(e.entry_offset, e.entry_length, cu.dwarfinfo.get_addr(cu, e.index)),  # type: ignore[has-type]
    'DW_LLE_startx_endx'     : lambda e, cu: LocationEntry(e.entry_offset, e.entry_length, cu.dwarfinfo.get_addr(cu, e.start_index), cu.dwarfinfo.get_addr(cu, e.end_index), e.loc_expr, True),  # type: ignore[has-type]
    'DW_LLE_startx_length'   : _translate_startx_length
}

class LocationListsPair:
    """For those binaries that contain both a debug_loc and a debug_loclists section,
    it holds a LocationLists object for both and forwards API calls to the right one.
    """
    def __init__(
        self,
        streamv4: IO[bytes],
        streamv5: IO[bytes],
        structs: DWARFStructs,
        dwarfinfo: DWARFInfo | None = None,
    ) -> None:
        self._loc = LocationLists(streamv4, structs, 4, dwarfinfo)
        self._loclists = LocationLists(streamv5, structs, 5, dwarfinfo)

    def get_location_list_at_offset(self, offset: int, die: DIE | None = None) -> list[_Location]:
        """See LocationLists.get_location_list_at_offset().
        """
        if die is None:
            raise DWARFError("For this binary, \"die\" needs to be provided")
        section = self._loclists if die.cu.header.version >= 5 else self._loc
        return section.get_location_list_at_offset(offset, die)

    def iter_location_lists(self) -> Iterator[BaseAddressEntry | LocationEntry]:
        """Tricky proposition, since the structure of loc and loclists
        is not identical. A realistic readelf implementation needs to be aware of both
        """
        raise DWARFError("Iterating through two sections is not supported")

    def iter_CUs(self) -> Iterator[CompileUnit]:
        """See LocationLists.iter_CUs()

        There are no CUs in DWARFv4 sections.
        """
        raise DWARFError("Iterating through two sections is not supported")

class LocationLists:
    """ A single location list is a Python list consisting of LocationEntry or
        BaseAddressEntry objects.

        Starting with DWARF5, it may also contain LocationViewPair, but only
        if scanning the section, never when requested for a DIE attribute.

        The default location entries are returned as LocationEntry with
        begin_offset == end_offset == -1

        Version determines whether the executable contains a debug_loc
        section, or a DWARFv5 style debug_loclists one. Only the 4/5
        distinction matters.

        Dwarfinfo is only needed for DWARFv5 location entry encodings
        that contain references to other sections (e. g. DW_LLE_startx_endx),
        and only for location list enumeration.
    """
    def __init__(
        self,
        stream: IO[bytes],
        structs: DWARFStructs,
        version: int = 4,
        dwarfinfo: DWARFInfo | None = None,
    ) -> None:
        self.stream = stream
        self.structs = structs
        self.dwarfinfo = dwarfinfo
        self.version = version
        self._max_addr: int = 2 ** (self.structs.address_size * 8) - 1

    def get_location_list_at_offset(self, offset: int, die: DIE | None = None) -> list[_Location]:
        """ Get a location list at the given offset in the section.
        Passing the die is only neccessary in DWARF5+, for decoding
        location entry encodings that contain references to other sections.
        """
        self.stream.seek(offset, os.SEEK_SET)
        if self.version >= 5:
            if die is None:
                raise DWARFError("For this binary, \"die\" needs to be provided")
            return self._parse_location_list_from_stream_v5(die.cu)
        return self._parse_location_list_from_stream()

    def iter_location_lists(self) -> Iterator[list[_Location]]:
        """ Iterates through location lists and view pairs. Returns lists of
        LocationEntry, BaseAddressEntry, and LocationViewPair objects.
        """
        # The location lists section was never meant for sequential access.
        # Location lists are referenced by DIE attributes by offset or by index.

        # As of DWARFv5, it may contain, in addition to proper location lists,
        # location list view pairs, which are referenced by the nonstandard DW_AT_GNU_locviews
        # attribute. A set of locview pairs (which is a couple of ULEB128 values) may preceed
        # a location list; the former is referenced by the DW_AT_GNU_locviews attribute, the
        # latter - by DW_AT_location (in the same DIE). Binutils' readelf dumps those.
        # There is a view pair for each location-type entry in the list.
        #
        # Also, the section may contain gaps.
        #
        # Taking a cue from binutils, we would have to scan this section while looking at
        # what's in DIEs.
        ver5 = self.version >= 5
        stream = self.stream
        stream.seek(0, os.SEEK_END)
        endpos = stream.tell()

        stream.seek(0, os.SEEK_SET)

        # Need to provide support for DW_AT_GNU_locviews. They are interspersed in
        # the locations section, no way to tell where short of checking all DIEs
        all_offsets = set() # Set of offsets where either a locview pair set can be found, or a view-less loclist
        locviews = dict() # Map of locview offset to the respective loclist offset
        cu_map = dict() # Map of loclist offsets to CUs
        assert self.dwarfinfo is not None
        for cu in self.dwarfinfo.iter_CUs():
            cu_ver: int = cu['version']
            if (cu_ver >= 5) == ver5:
                for die in cu.iter_DIEs():
                    # A combination of location and locviews means there is a location list
                    # preceed by several locview pairs
                    if 'DW_AT_GNU_locviews' in die.attributes:
                        assert('DW_AT_location' in die.attributes and
                            LocationParser._attribute_has_loc_list(die.attributes['DW_AT_location'], cu_ver))
                        views_offset: int = die.attributes['DW_AT_GNU_locviews'].value
                        list_offset: int = die.attributes['DW_AT_location'].value
                        locviews[views_offset] = list_offset
                        cu_map[list_offset] = cu
                        all_offsets.add(views_offset)

                    # Scan other attributes for location lists
                    for key in die.attributes:
                        attr = die.attributes[key]
                        if ((key != 'DW_AT_location' or 'DW_AT_GNU_locviews' not in die.attributes) and
                            LocationParser.attribute_has_location(attr, cu_ver) and
                            LocationParser._attribute_has_loc_list(attr, cu_ver)):
                            list_offset = attr.value
                            all_offsets.add(list_offset)
                            cu_map[list_offset] = cu
        sorted_offsets = sorted(all_offsets)

        if ver5:
            # Loclists section is organized as an array of CUs, each length prefixed.
            # We don't assume that the CUs go in the same order as the ones in info.
            offset_index = 0
            while stream.tell() < endpos:
                # We are at the start of the CU block in the loclists now
                cu_header = struct_parse(self.structs.Dwarf_loclists_CU_header, stream)
                assert(cu_header.version == 5)

                # GNU binutils supports two traversal modes: by offsets in CU header, and sequential.
                # We don't have a binary for the former yet. On an off chance that we one day might,
                # let's parse the header anyway.

                cu_end_offset: int = cu_header.offset_after_length + cu_header.unit_length
                # Unit_length includes the header but doesn't include the length

                while stream.tell() < cu_end_offset:
                    # Skip the gap to the next object
                    next_offset = sorted_offsets[offset_index]
                    if next_offset == stream.tell(): # At an object, either a loc list or a loc view pair
                        locview_pairs = self._parse_locview_pairs(locviews)
                        entries = self._parse_location_list_from_stream_v5(cu_map[stream.tell()])
                        yield locview_pairs + entries
                        offset_index += 1
                    else: # We are at a gap - skip the gap to the next object or to the next CU
                        if next_offset > cu_end_offset: # Gap at the CU end - the next object is in the next CU
                            next_offset = cu_end_offset # And implicitly quit the loop within the CU
                        stream.seek(next_offset, os.SEEK_SET)
        else:
            for offset in sorted_offsets:
                list_offset = locviews.get(offset, offset)
                if cu_map[list_offset].header.version < 5:
                    stream.seek(offset, os.SEEK_SET)
                    locview_pairs = self._parse_locview_pairs(locviews)
                    entries = self._parse_location_list_from_stream()
                    yield locview_pairs + entries

    def iter_CUs(self) -> Iterator[CompileUnit]:
        """For DWARF5 returns an array of objects, where each one has an array of offsets
        """
        if self.version < 5:
            raise DWARFError("CU iteration in loclists is not supported with DWARF<5")

        assert self.dwarfinfo is not None
        structs = next(self.dwarfinfo.iter_CUs()).structs # Just pick one
        return _iter_CUs_in_section(self.stream, structs, structs.Dwarf_loclists_CU_header)

    #------ PRIVATE ------#

    def _parse_location_list_from_stream(self) -> list[_Location]:
        lst: list[_Location] = []
        while True:
            entry_offset = self.stream.tell()
            begin_offset: int = struct_parse(
                self.structs.the_Dwarf_target_addr, self.stream)
            end_offset: int = struct_parse(
                self.structs.the_Dwarf_target_addr, self.stream)
            if begin_offset == 0 and end_offset == 0:
                # End of list - we're done.
                break
            elif begin_offset == self._max_addr:
                # Base address selection entry
                entry_length = self.stream.tell() - entry_offset
                lst.append(BaseAddressEntry(entry_offset=entry_offset, entry_length=entry_length, base_address=end_offset))
            else:
                # Location list entry
                expr_len: int = struct_parse(
                    self.structs.the_Dwarf_uint16, self.stream)
                loc_expr: list[int] = [struct_parse(self.structs.the_Dwarf_uint8,
                                         self.stream)
                                for i in range(expr_len)]
                entry_length = self.stream.tell() - entry_offset
                lst.append(LocationEntry(
                    entry_offset=entry_offset,
                    entry_length=entry_length,
                    begin_offset=begin_offset,
                    end_offset=end_offset,
                    loc_expr=loc_expr,
                    is_absolute = False))
        return lst

    def _parse_location_list_from_stream_v5(
        self,
        cu: CompileUnit | TypeUnit | None = None,
    ) -> list[_Location]:
        """ Returns an array with BaseAddressEntry and LocationEntry.
            No terminator entries.

            The cu argument is necessary if the section is a
            DWARFv5 debug_loclists one, and the target loclist
            contains indirect encodings.
        """
        return [entry_translate[entry.entry_type](entry, cu)  # type: ignore[arg-type] # ty: ignore[invalid-argument-type]
            for entry
            in struct_parse(self.structs.Dwarf_loclists_entries, self.stream)]

    # From V5 style entries to a LocationEntry/BaseAddressEntry
    def _translate_entry_v5(self, entry: Container, die: DIE) -> _Location:
        off: int = entry.entry_offset
        len: int = entry.entry_end_offset - off
        type: str = entry.entry_type
        if type == 'DW_LLE_base_address':
            return BaseAddressEntry(off, len, entry.address)
        elif type == 'DW_LLE_offset_pair':
            return LocationEntry(off, len, entry.start_offset, entry.end_offset, entry.loc_expr, False)
        elif type == 'DW_LLE_start_length':
            return LocationEntry(off, len, entry.start_address, entry.start_address + entry.length, entry.loc_expr, True)
        elif type == 'DW_LLE_start_end': # No test for this yet, but the format seems straightforward
            return LocationEntry(off, len, entry.start_address, entry.end_address, entry.loc_expr, True)
        elif type == 'DW_LLE_default_location': # No test for this either, and this is new in the API
            return LocationEntry(off, len, -1, -1, entry.loc_expr, True)
        elif type in ('DW_LLE_base_addressx', 'DW_LLE_startx_endx', 'DW_LLE_startx_length'):
            # We don't have sample binaries for those LLEs. Their proper parsing would
            # require knowing the CU context (so that indices can be resolved to code offsets)
            raise NotImplementedError("Location list entry type %s is not supported yet" % (type,))
        else:
            raise DWARFError(False, "Unknown DW_LLE code: %s" % (type,))

    # Locviews is the dict, mapping locview offsets to corresponding loclist offsets
    def _parse_locview_pairs(self, locviews: Mapping[int, int]) -> list[LocationViewPair]:
        stream = self.stream
        list_offset: int | None = locviews.get(stream.tell(), None)
        pairs: list[LocationViewPair] = []
        if list_offset is not None:
            while stream.tell() < list_offset:
                pair = struct_parse(self.structs.Dwarf_locview_pair, stream)
                pairs.append(LocationViewPair(pair.entry_offset, pair.begin, pair.end))
            assert(stream.tell() == list_offset)
        return pairs

class LocationParser:
    """ A parser for location information in DIEs.
        Handles both location information contained within the attribute
        itself (represented as a LocationExpr object) and references to
        location lists in the .debug_loc section (represented as a
        list).
    """
    def __init__(self, location_lists: LocationLists | LocationListsPair | None) -> None:
        self.location_lists = location_lists

    @staticmethod
    def attribute_has_location(attr: AttributeValue, dwarf_version: int) -> bool:
        """ Checks if a DIE attribute contains location information.
        """
        return (LocationParser._attribute_is_loclistptr_class(attr) and
                (LocationParser._attribute_has_loc_expr(attr, dwarf_version) or
                 LocationParser._attribute_has_loc_list(attr, dwarf_version)))

    def parse_from_attribute(
        self,
        attr: AttributeValue,
        dwarf_version: int,
        die: DIE | None = None,
    ) -> LocationExpr | list[_Location]:
        """ Parses a DIE attribute and returns either a LocationExpr or
            a list.
        """
        if self.attribute_has_location(attr, dwarf_version):
            if self._attribute_has_loc_expr(attr, dwarf_version):
                return LocationExpr(attr.value)
            elif self._attribute_has_loc_list(attr, dwarf_version):
                assert self.location_lists is not None
                return self.location_lists.get_location_list_at_offset(
                    attr.value, die)
                # We don't yet know if the DIE context will be needed.
                # We might get it without a full tree traversal using
                # attr.offset as a key, but we assume a good DWARF5
                # aware consumer would pass a DIE along.
        raise ValueError("Attribute does not have location information")

    #------ PRIVATE ------#

    @staticmethod
    def _attribute_has_loc_expr(attr: AttributeValue, dwarf_version: int) -> bool:
        return ((dwarf_version < 4 and attr.form.startswith('DW_FORM_block') and
            not attr.name == 'DW_AT_const_value') or
            attr.form == 'DW_FORM_exprloc')

    @staticmethod
    def _attribute_has_loc_list(attr: AttributeValue, dwarf_version: int) -> bool:
        return (((dwarf_version < 4 and
                 attr.form in ('DW_FORM_data1', 'DW_FORM_data2', 'DW_FORM_data4', 'DW_FORM_data8') and
                 not attr.name == 'DW_AT_const_value') or
                attr.form in ('DW_FORM_sec_offset', 'DW_FORM_loclistx')) and
                not LocationParser._attribute_is_constant(attr, dwarf_version))

    # Starting with DWARF3, DW_AT_data_member_location may contain an integer offset
    # instead of a location expression. Need to prevent false positives on attribute_has_location().
    # As for DW_AT_upper_bound/DW_AT_count, we've seen it in form DW_FORM_locexpr in a V5 binary. usually it's a constant,
    # but the constant sholdn't be misinterpreted as a loclist pointer.
    @staticmethod
    def _attribute_is_constant(attr: AttributeValue, dwarf_version: int) -> bool:
        return (((dwarf_version >= 3 and attr.name == 'DW_AT_data_member_location') or
                 (attr.name in ('DW_AT_upper_bound', 'DW_AT_count'))) and
            attr.form in ('DW_FORM_data1', 'DW_FORM_data2', 'DW_FORM_data4', 'DW_FORM_data8', 'DW_FORM_sdata', 'DW_FORM_udata'))

    @staticmethod
    def _attribute_is_loclistptr_class(attr: AttributeValue) -> bool:
        return (attr.name in ( 'DW_AT_location', 'DW_AT_string_length',
                               'DW_AT_const_value', 'DW_AT_return_addr',
                               'DW_AT_data_member_location',
                               'DW_AT_frame_base', 'DW_AT_segment',
                               'DW_AT_static_link', 'DW_AT_use_location',
                               'DW_AT_vtable_elem_location',
                               'DW_AT_call_value',
                               'DW_AT_GNU_call_site_value',
                               'DW_AT_GNU_call_site_target',
                               'DW_AT_GNU_call_site_data_value',
                               'DW_AT_call_target',
                               'DW_AT_call_target_clobbered',
                               'DW_AT_call_data_location',
                               'DW_AT_call_data_value',
                               'DW_AT_upper_bound',
                               'DW_AT_count'))



#-------------------------------------------------------------------------------
# elftools: elf/dynamic.py
#
# ELF Dynamic Tags
#
# Mike Frysinger (vapier@gentoo.org)
# This code is in the public domain
#-------------------------------------------------------------------------------
from __future__ import annotations

import itertools
from collections import defaultdict
from functools import cached_property
from typing import IO, TYPE_CHECKING, Any, Protocol, TypedDict, cast, runtime_checkable

from ..common.exceptions import ELFError
from ..common.utils import elf_assert, struct_parse, parse_cstring_from_stream
from .enums import ENUM_D_TAG
from .hash import ELFHashTable, GNUHashTable
from .relocation import RelocationTable, RelrRelocationTable
from .sections import Section, Symbol
from .segments import Segment

if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..construct.lib.container import Container
    from .elffile import ELFFile


class RelocationTables(TypedDict, total=False):
    REL: RelocationTable
    RELA: RelocationTable
    RELR: RelrRelocationTable
    JMPREL: RelocationTable


@runtime_checkable
class _StringTable(Protocol):
    """Common base-class of elftools.elf.dynamic._DynamicStringTable and
    elftools.elf.section.StringTableSection to be consumed by
    DynamicTag|Dynamic.
    Requires @runtime_checkable as `assert isinstance(…, _StringTable)` is
    used."""
    def get_string(self, offset: int, /) -> str: ...


class _DynamicStringTable:
    """ Bare string table based on values found via ELF dynamic tags and
        loadable segments only.  Good enough for get_string() only.
    """
    def __init__(self, stream: IO[bytes], table_offset: int) -> None:
        self._stream = stream
        self._table_offset = table_offset

    def get_string(self, offset: int) -> str:
        """ Get the string stored at the given offset in this string table.
        """
        s = parse_cstring_from_stream(self._stream, self._table_offset + offset)
        return s.decode('utf-8') if s else ''


class DynamicTag:
    """ Dynamic Tag object - representing a single dynamic tag entry from a
        dynamic section.

        Allows dictionary-like access to the dynamic structure. For special
        tags (those listed in the _HANDLED_TAGS set below), creates additional
        attributes for convenience. For example, .soname will contain the actual
        value of DT_SONAME (fetched from the dynamic symbol table).
    """
    _HANDLED_TAGS = frozenset(
        ['DT_NEEDED', 'DT_RPATH', 'DT_RUNPATH', 'DT_SONAME',
         'DT_SUNW_FILTER'])

    def __init__(
        self,
        entry: Container,
        stringtable: _StringTable | None,
    ) -> None:
        if stringtable is None:
            raise ELFError('Creating DynamicTag without string table')
        self.entry = entry
        if entry.d_tag in self._HANDLED_TAGS:
            setattr(self, entry.d_tag[3:].lower(),
                    stringtable.get_string(self.entry.d_val))

    def __getitem__(self, name: str) -> Any:
        """ Implement dict-like access to entries
        """
        return self.entry[name]

    def __repr__(self) -> str:
        return '<DynamicTag (%s): %r>' % (self.entry.d_tag, self.entry)

    def __str__(self) -> str:
        if self.entry.d_tag in self._HANDLED_TAGS:
            s = '"%s"' % getattr(self, self.entry.d_tag[3:].lower())
        else:
            s = '%#x' % self.entry.d_ptr
        return '<DynamicTag (%s) %s>' % (self.entry.d_tag, s)


class Dynamic:
    """ Shared functionality between dynamic sections and segments.
    """
    def __init__(
        self,
        stream: IO[bytes],
        elffile: ELFFile,
        stringtable: _StringTable | Section | None,
        position: int,
        empty: bool,
    ) -> None:
        """
        stream:
            The file-like object from which to load data

        elffile:
            The parent elffile object

        stringtable:
            A stringtable reference to use for parsing string references in
            entries

        position:
            The file offset of the dynamic segment/section

        empty:
            Whether this is a degenerate case with zero entries. Normally, every
            dynamic table will have at least one entry, the DT_NULL terminator.
        """
        self.elffile = elffile
        self.elfstructs = elffile.structs
        self._stream = stream
        self._num_tags = -1 if not empty else 0
        self._offset = position
        self._tagsize = self.elfstructs.Elf_Dyn.sizeof()
        self._empty = empty

        # Do not access this directly yourself; use _get_stringtable() instead.
        self._stringtable: _StringTable | Section | None = stringtable

    def get_table_offset(self, tag_name: str) -> tuple[int | None, int | None]:
        """ Return the virtual address and file offset of a dynamic table.
        """
        try:
            ptr: int = next(
                tag['d_ptr']
                for tag in self._iter_tags(type=tag_name)
            )
        except StopIteration:
            return (None, None)

        # If we found a virtual address, locate the offset in the file
        # by using the program headers.
        offset = next(self.elffile.address_offsets(ptr), None)

        return ptr, offset

    def _get_stringtable(self) -> _StringTable:
        """ Return a string table for looking up dynamic tag related strings.

            This won't be a "full" string table object, but will at least
            support the get_string() function.
        """
        if self._stringtable:
            assert isinstance(self._stringtable, _StringTable)
            return self._stringtable

        # If the ELF has stripped its section table (which is unusual, but
        # perfectly valid), we need to use the dynamic tags to locate the
        # dynamic string table.
        _, table_offset = self.get_table_offset('DT_STRTAB')
        if table_offset is not None:
            self._stringtable = _DynamicStringTable(self._stream, table_offset)
            assert isinstance(self._stringtable, _StringTable)
            return self._stringtable

        # That didn't work for some reason.  Let's use the section header
        # even though this ELF is super weird.
        self._stringtable = self.elffile.get_section_by_name('.dynstr')
        assert isinstance(self._stringtable, _StringTable)
        return self._stringtable

    def _iter_tags(self, type: str | None = None) -> Iterator[Container]:
        """ Yield all raw tags (limit to |type| if specified)
        """
        if self._empty:
            return
        for n in itertools.count():
            tag = self._get_tag(n)
            if type is None or tag['d_tag'] == type:
                yield tag
            if tag['d_tag'] == 'DT_NULL':
                break

    def iter_tags(self, type: str | None = None) -> Iterator[DynamicTag]:
        """ Yield all tags (limit to |type| if specified)
        """
        for tag in self._iter_tags(type=type):
            yield DynamicTag(tag, self._get_stringtable())

    def _get_tag(self, n: int) -> Container:
        """ Get the raw tag at index #n from the file
        """
        if self._num_tags != -1 and n >= self._num_tags:
            raise IndexError(n)
        offset = self._offset + n * self._tagsize
        return struct_parse(
            self.elfstructs.Elf_Dyn,
            self._stream,
            stream_pos=offset)

    def get_tag(self, n: int) -> DynamicTag:
        """ Get the tag at index #n from the file (DynamicTag object)
        """
        return DynamicTag(self._get_tag(n), self._get_stringtable())

    def num_tags(self) -> int | None:
        """ Number of dynamic tags in the file, including the DT_NULL tag
        """
        if self._num_tags != -1:
            return self._num_tags

        for n in itertools.count():
            tag = self.get_tag(n)
            if tag.entry.d_tag == 'DT_NULL':
                self._num_tags = n + 1
                return self._num_tags

        return None

    def get_relocation_tables(self) -> RelocationTables:
        """ Load all available relocation tables from DYNAMIC tags.

            Returns a dictionary mapping found table types (REL, RELA,
            RELR, JMPREL) to RelocationTable objects.
        """

        result: RelocationTables = {}

        if list(self.iter_tags('DT_REL')):
            result['REL'] = RelocationTable(self.elffile,
                self.get_table_offset('DT_REL')[1],  # type: ignore[arg-type] # ty: ignore[invalid-argument-type]
                next(self.iter_tags('DT_RELSZ'))['d_val'], False)

            relentsz = next(self.iter_tags('DT_RELENT'))['d_val']
            elf_assert(result['REL'].entry_size == relentsz,
                'Expected DT_RELENT to be %s' % relentsz)

        if list(self.iter_tags('DT_RELA')):
            result['RELA'] = RelocationTable(self.elffile,
                self.get_table_offset('DT_RELA')[1],  # type: ignore[arg-type] # ty: ignore[invalid-argument-type]
                next(self.iter_tags('DT_RELASZ'))['d_val'], True)

            relentsz = next(self.iter_tags('DT_RELAENT'))['d_val']
            elf_assert(result['RELA'].entry_size == relentsz,
                'Expected DT_RELAENT to be %s' % relentsz)

        if list(self.iter_tags('DT_RELR')):
            result['RELR'] = RelrRelocationTable(self.elffile,
                self.get_table_offset('DT_RELR')[1],  # type: ignore[arg-type] # ty: ignore[invalid-argument-type]
                next(self.iter_tags('DT_RELRSZ'))['d_val'],
                next(self.iter_tags('DT_RELRENT'))['d_val'])

        if list(self.iter_tags('DT_JMPREL')):
            result['JMPREL'] = RelocationTable(self.elffile,
                self.get_table_offset('DT_JMPREL')[1],  # type: ignore[arg-type] # ty: ignore[invalid-argument-type]
                next(self.iter_tags('DT_PLTRELSZ'))['d_val'],
                next(self.iter_tags('DT_PLTREL'))['d_val'] == ENUM_D_TAG['DT_RELA'])

        return result


class DynamicSection(Section, Dynamic):
    """ ELF dynamic table section.  Knows how to process the list of tags.
    """
    def __init__(self, header: Container, name: str, elffile: ELFFile) -> None:
        Section.__init__(self, header, name, elffile)
        stringtable = elffile.get_section(header['sh_link'], ('SHT_STRTAB', 'SHT_NOBITS', 'SHT_NULL'))
        Dynamic.__init__(self, self.stream, self.elffile, stringtable,
            self['sh_offset'], self['sh_type'] == 'SHT_NOBITS')


class DynamicSegment(Segment, Dynamic):
    """ ELF dynamic table segment.  Knows how to process the list of tags.
    """
    def __init__(self, header: Container, stream: IO[bytes], elffile: ELFFile) -> None:
        # The string table section to be used to resolve string names in
        # the dynamic tag array is the one pointed at by the sh_link field
        # of the dynamic section header.
        # So we must look for the dynamic section contained in the dynamic
        # segment, we do so by searching for the dynamic section whose content
        # is located at the same offset as the dynamic segment
        stringtable = next(
            (
                elffile.get_section(section['sh_link'])
                for section in elffile.iter_sections()
                if isinstance(section, DynamicSection) and section['sh_offset'] == header['p_offset']
            ),
            None,
        )
        Segment.__init__(self, header, stream)
        Dynamic.__init__(self, stream, elffile, stringtable, self['p_offset'],
             self['p_filesz'] == 0)
        self._symbol_size = self.elfstructs.Elf_Sym.sizeof()

    def num_symbols(self) -> int:
        """ Number of symbols in the table recovered from DT_SYMTAB
        """
        return self._num_symbols

    @cached_property
    def _num_symbols(self) -> int:
        # Check if a DT_GNU_HASH tag exists and recover the number of symbols
        # from the corresponding hash table
        _, gnu_hash_offset = self.get_table_offset('DT_GNU_HASH')
        if gnu_hash_offset is not None:
            gnu_hash_section = GNUHashTable(self.elffile, gnu_hash_offset, self)
            return gnu_hash_section.get_number_of_symbols()

        # If DT_GNU_HASH did not exist, maybe we can use DT_HASH
        _, hash_offset = self.get_table_offset('DT_HASH')
        if hash_offset is not None:
            # Get the hash table from the DT_HASH offset
            hash_section = ELFHashTable(self.elffile, hash_offset, None, self)
            return hash_section.get_number_of_symbols()

        # Find closest higher pointer than tab_ptr. We'll use that to mark
        # the end of the symbol table.
        tab_ptr, tab_offset = self.get_table_offset('DT_SYMTAB')
        if tab_ptr is None or tab_offset is None:
            raise ELFError('Segment does not contain DT_SYMTAB.')

        nearest_ptr: int | None = None
        for tag in self.iter_tags():
            tag_ptr = tag['d_ptr']
            if tag['d_tag'] == 'DT_SYMENT':
                if self._symbol_size != tag['d_val']:
                    # DT_SYMENT is the size of one symbol entry. It must be
                    # the same as returned by Elf_Sym.sizeof.
                    raise ELFError('DT_SYMENT (%d) != Elf_Sym (%d).' %
                                   (tag['d_val'], self._symbol_size))
            if (tag_ptr > tab_ptr and
                    (nearest_ptr is None or nearest_ptr > tag_ptr)):
                nearest_ptr = tag_ptr

        if nearest_ptr is not None:
            return (nearest_ptr - tab_ptr) // self._symbol_size

        # Use the end of last segment that contains DT_SYMTAB (or ends on it)
        for segment in self.elffile.iter_segments(type='PT_LOAD'):
            start = segment['p_vaddr']
            end = start + segment['p_filesz']
            if start <= tab_ptr <= end:
                nearest_ptr = end

        if nearest_ptr is not None:
            return (nearest_ptr - tab_ptr) // self._symbol_size

        raise ELFError('Cannot determine the end of DT_SYMTAB.')

    def get_symbol(self, index: int) -> Symbol:
        """ Get the symbol at index #index from the table (Symbol object)
        """
        tab_ptr, tab_offset = self.get_table_offset('DT_SYMTAB')
        if tab_ptr is None or tab_offset is None:
            raise ELFError('Segment does not contain DT_SYMTAB.')

        symbol = struct_parse(
            self.elfstructs.Elf_Sym,
            self._stream,
            stream_pos=tab_offset + index * self._symbol_size)

        string_table = self._get_stringtable()
        symbol_name = string_table.get_string(symbol["st_name"])

        return Symbol(symbol, symbol_name)

    def get_symbol_by_name(self, name: str) -> list[Symbol] | None:
        """ Get a symbol(s) by name. Return None if no symbol by the given name
            exists.
        """
        symnums = self._symbol_name_map.get(name)
        return [self.get_symbol(i) for i in symnums] if symnums else None

    @cached_property
    def _symbol_name_map(self) -> dict[str, list[int]]:
        smap = defaultdict(list)
        for i, sym in enumerate(self.iter_symbols()):
            smap[sym.name].append(i)
        return smap

    def iter_symbols(self) -> Iterator[Symbol]:
        """ Yield all symbols in this dynamic segment. The symbols are usually
            the same as returned by SymbolTableSection.iter_symbols. However,
            in stripped binaries, SymbolTableSection might have been removed.
            This method reads from the mandatory dynamic tag DT_SYMTAB.
        """
        for i in range(self.num_symbols()):
            yield self.get_symbol(i)



"""
Various containers.
"""
from __future__ import annotations

from collections.abc import MutableMapping
from functools import wraps
from pprint import pformat
from typing import IO, TYPE_CHECKING, Any, Literal, overload

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from typing import Concatenate, ParamSpec, TypeVar

    from typing_extensions import Self  # 3.11+

    from ..core import Construct
    from .hex import HexString

    _P = ParamSpec('_P')
    _R = TypeVar('_R')
    _T = TypeVar('_T')


__all__ = [
    "recursion_lock",
    "Container", "FlagsContainer", "ListContainer", "LazyContainer",
]


def recursion_lock(
    retval: _R,
    lock_name: str = "__recursion_lock__",
) -> Callable[[Callable[Concatenate[Any, _P], _T]], Callable[Concatenate[Any, _P], _T | _R]]:

    def decorator(
        func: Callable[Concatenate[Any, _P], _T],
    ) -> Callable[Concatenate[Any, _P], _T | _R]:
        @wraps(func)
        def wrapper(self: Any, *args: _P.args, **kw: _P.kwargs) -> _T | _R:
            if getattr(self, lock_name, False):
                return retval
            setattr(self, lock_name, True)
            try:
                return func(self, *args, **kw)
            finally:
                setattr(self, lock_name, False)
        return wrapper
    return decorator

class Container(MutableMapping[str, Any]):
    """
    A generic container of attributes.

    Containers are the common way to express parsed data.
    """

    def __init__(self, **kw: Any) -> None:
        self.__dict__ = kw

    # The core dictionary interface.

    @overload
    def __getitem__(self, name: Literal[
        "ch_addralign", "ch_size",
        "length",
        "n_descsz", "n_offset", "n_namesz",
        "sh_addralign", "sh_flags", "sh_size",
        "bloom_size", "nbuckets", "nchains",
    ]) -> int: ...
    @overload
    def __getitem__(self, name: Literal[
        "ch_type",
        "sh_type",
        "n_name", "n_type",
        "tag", "vendor_name",
    ]) -> str: ...
    @overload
    def __getitem__(self, name: Literal[
        "buckets", "chains",
    ]) -> list[int]: ...
    @overload
    def __getitem__(self, name: str) -> Any: ...
    def __getitem__(self, name: str) -> Any:
        return self.__dict__[name]

    def __delitem__(self, name: str) -> None:
        del self.__dict__[name]

    def __setitem__(self, name: str, value: Any) -> None:
        self.__dict__[name] = value

    def __iter__(self) -> Iterator[str]:
        return iter(self.__dict__)

    def __len__(self) -> int:
        return len(self.__dict__.keys())

    # Copy interface.

    def copy(self) -> Self:
        return self.__class__(**self.__dict__)

    __copy__ = copy

    def __repr__(self) -> str:
        return "%s(%s)" % (self.__class__.__name__, repr(self.__dict__))

    def __str__(self) -> str:
        return "%s(%s)" % (self.__class__.__name__, str(self.__dict__))

    if TYPE_CHECKING:
        # elftools.construct.debug Probe.printout()
        stream_position: int
        following_stream_data: str | HexString
        context: Container
        stack: ListContainer
        # allow arbitray attributes
        def __setattr__(self, name: str, value: object) -> None: ...
        def __getattr__(self, name: str) -> Any: ...

class FlagsContainer(Container):
    """
    A container providing pretty-printing for flags.

    Only set flags are displayed.
    """

    @recursion_lock("<...>")
    def __str__(self) -> str:
        d = dict((k, self[k]) for k in self
                 if self[k] and not k.startswith("_"))
        return "%s(%s)" % (self.__class__.__name__, pformat(d))

class ListContainer(list[Any]):
    """
    A container for lists.
    """

    __slots__ = ("__recursion_lock__",)

    @recursion_lock("[...]")
    def __str__(self) -> str:
        return pformat(self)

class LazyContainer:

    __slots__ = ("subcon", "stream", "pos", "context", "_value")

    def __init__(self, subcon: Construct, stream: IO[bytes], pos: int, context: Container) -> None:
        self.subcon = subcon
        self.stream = stream
        self.pos = pos
        self.context = context
        self._value = NotImplemented

    def __eq__(self, other: object) -> bool:
        return isinstance(other, LazyContainer) and self._value == other._value

    def __ne__(self, other: object) -> bool:
        return not (self == other)

    def __str__(self) -> str:
        return self.__pretty_str__()

    def __pretty_str__(self, nesting: int = 1, indentation: str = "    ") -> str:
        if self._value is NotImplemented:
            text = "<unread>"
        elif hasattr(self._value, "__pretty_str__"):
            text = self._value.__pretty_str__(nesting, indentation)
        else:
            text = str(self._value)
        return "%s: %s" % (self.__class__.__name__, text)

    def read(self) -> Any:
        self.stream.seek(self.pos)
        return self.subcon._parse(self.stream, self.context)

    def dispose(self) -> None:
        del self.subcon
        del self.stream
        del self.context
        del self.pos

    def _get_value(self) -> Any:
        if self._value is NotImplemented:
            self._value = self.read()
        return self._value

    value = property(_get_value)

    has_value = property(lambda self: self._value is not NotImplemented)



#-------------------------------------------------------------------------------
# elftools: dwarf/constants.py
#
# Constants and flags
#
# Eli Bendersky (eliben@gmail.com)
# This code is in the public domain
#-------------------------------------------------------------------------------
from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Self  # 3.11+


class _IntEnum(int, Enum):  # Py3.11: enum.ReprEnum
    def __repr__(self) -> str:
        return int.__str__(self.value)

    @property  # Py3.11+: enum.property
    def FQN(self) -> str:
        return f"{self.__class__.__name__}_{self.name}"


class DW_INL(_IntEnum):
    """Inline codes."""
    not_inlined = 0
    inlined = 1
    declared_not_inlined = 2
    declared_inlined = 3


class DW_LANG(_IntEnum):
    """Source languages."""
    C89 = 0x0001
    C = 0x0002
    Ada83 = 0x0003
    C_plus_plus = 0x0004
    Cobol74 = 0x0005
    Cobol85 = 0x0006
    Fortran77 = 0x0007
    Fortran90 = 0x0008
    Pascal83 = 0x0009
    Modula2 = 0x000a
    Java = 0x000b
    C99 = 0x000c
    Ada95 = 0x000d
    Fortran95 = 0x000e
    PLI = 0x000f
    ObjC = 0x0010
    ObjC_plus_plus = 0x0011
    UPC = 0x0012
    D = 0x0013
    Python = 0x0014
    OpenCL = 0x0015
    Go = 0x0016
    Modula3 = 0x0017
    Haskell = 0x0018
    C_plus_plus_03 = 0x0019
    C_plus_plus_11 = 0x001a
    OCaml = 0x001b
    Rust = 0x001c
    C11 = 0x001d
    Swift = 0x001e
    Julia = 0x001f
    Dylan = 0x0020
    C_plus_plus_14 = 0x0021
    Fortran03 = 0x0022
    Fortran08 = 0x0023
    RenderScript = 0x0024
    BLISS = 0x0025
    Mips_Assembler = 0x8001
    Upc = 0x8765
    HP_Bliss = 0x8003
    HP_Basic91 = 0x8004
    HP_Pascal91 = 0x8005
    HP_IMacro = 0x8006
    HP_Assembler = 0x8007
    GOOGLE_RenderScript = 0x8e57
    BORLAND_Delphi = 0xb000


class DW_ATE(_IntEnum):
    """Encodings."""
    void = 0x0
    address = 0x1
    boolean = 0x2
    complex_float = 0x3
    float = 0x4
    signed = 0x5
    signed_char = 0x6
    unsigned = 0x7
    unsigned_char = 0x8
    imaginary_float = 0x9
    packed_decimal = 0xa
    numeric_string = 0xb
    edited = 0xc
    signed_fixed = 0xd
    unsigned_fixed = 0xe
    decimal_float = 0xf
    UTF = 0x10
    UCS = 0x11
    ASCII = 0x12
    lo_user = 0x80
    hi_user = 0xff
    HP_float80 = 0x80
    HP_complex_float80 = 0x81
    HP_float128 = 0x82
    HP_complex_float128 = 0x83
    HP_floathpintel = 0x84
    HP_imaginary_float80 = 0x85
    HP_imaginary_float128 = 0x86


class DW_ACCESS(_IntEnum):
    """Access."""
    public = 1
    protected = 2
    private = 3


class DW_VIS(_IntEnum):
    """Visibility."""
    local = 1
    exported = 2
    qualified = 3


class DW_VIRTUALITY(_IntEnum):
    """Virtuality."""
    none = 0
    virtual = 1
    pure_virtual = 2


class DW_ID(_IntEnum):
    """ID cases."""
    case_sensitive = 0
    up_case = 1
    down_case = 2
    case_insensitive = 3


class DW_CC(_IntEnum):
    """Calling conventions."""
    normal = 0x1
    program = 0x2
    nocall = 0x3
    pass_by_reference = 0x4
    pass_by_valuee = 0x5


class DW_ORD(_IntEnum):
    """Orderings."""
    row_major = 0
    col_major = 1


class DW_LNS(_IntEnum):
    """Line program opcodes."""
    copy = 0x01
    advance_pc = 0x02
    advance_line = 0x03
    set_file = 0x04
    set_column = 0x05
    negate_stmt = 0x06
    set_basic_block = 0x07
    const_add_pc = 0x08
    fixed_advance_pc = 0x09
    set_prologue_end = 0x0a
    set_epilogue_begin = 0x0b
    set_isa = 0x0c


class DW_LNE(_IntEnum):
    """Line program extended opcodes."""
    end_sequence = 0x01
    set_address = 0x02
    define_file = 0x03
    set_discriminator = 0x04
    lo_user = 0x80
    hi_user = 0xff


class DW_LNCT(_IntEnum):
    """Line program header content types."""
    path = 0x01
    directory_index = 0x02
    timestamp = 0x03
    size = 0x04
    MD5 = 0x05
    lo_user = 0x2000
    LLVM_source = 0x2001
    LLVM_is_MD5 = 0x2002
    hi_user = 0x3fff


class DW_CFA(_IntEnum):
    """
    Call frame instructions.

    Note that the first 3 instructions have the so-called "primary opcode"
    (as described in DWARFv3 7.23), so only their highest 2 bits take part
    in the opcode decoding. They are kept as constants with the low bits masked
    out, and the callframe module knows how to handle this.
    The other instructions use an "extended opcode" encoded just in the low 6
    bits, with the high 2 bits, so these constants are exactly as they would
    appear in an actual file.
    """
    advance_loc = 0b01000000
    offset = 0b10000000
    restore = 0b11000000

    nop = 0x00
    set_loc = 0x01
    advance_loc1 = 0x02
    advance_loc2 = 0x03
    advance_loc4 = 0x04
    offset_extended = 0x05
    restore_extended = 0x06
    undefined = 0x07
    same_value = 0x08
    register = 0x09
    remember_state = 0x0a
    restore_state = 0x0b
    def_cfa = 0x0c
    def_cfa_register = 0x0d
    def_cfa_offset = 0x0e
    def_cfa_expression = 0x0f
    expression = 0x10
    offset_extended_sf = 0x11
    def_cfa_sf = 0x12
    def_cfa_offset_sf = 0x13
    val_offset = 0x14
    val_offset_sf = 0x15
    val_expression = 0x16
    AARCH64_negate_ra_state = 0x2d
    GNU_window_save = 0x2d  # Used on SPARC, not in the corpus
    GNU_args_size = 0x2e

    @classmethod
    def parse_raw_opcode(cls, /, opcode: int, *, __MASK: int = 0b11_00_0000) -> tuple[Self, int] | tuple[Self]:
        """Extract primary or extended opcode from raw byte."""
        if primary := opcode & __MASK:
            return (cls(primary), opcode & ~__MASK)
        return (cls(opcode),)


class DW_UT(_IntEnum):
    """
    Compilation unit types.

    DWARFv5 introduces the "unit_type" field to each CU header, allowing
    individual CUs to indicate whether they're complete, partial, and so forth.
    See DWARFv5 3.1 ("Unit Entries") and 7.5.1 ("Unit Headers").
    """
    compile = 0x01
    type = 0x02
    partial = 0x03
    skeleton = 0x04
    split_compile = 0x05
    split_type = 0x06
    lo_user = 0x80
    hi_user = 0xff


# Add back legacy names `DW_UT_type = DW_UT.type` for `from .constants import *`.
# These are invisible to typing as the members are added dynamically by code!
# Use __members__ to also add aliases like DW_CFA.{AARCH64_negate_ra_state,GNU_window_save}.
globals().update({
    f"{enum_name}_{member_name}": member.value
    for enum_name, enum in globals().items()
    if enum_name.startswith("DW_") and issubclass(enum, _IntEnum)
    for member_name, member in enum.__members__.items()
})



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Self  # 3.11+


# Map an integer in the inclusive range 0-255 to its string byte representation
_printable = {i: chr(i) if 32 <= i < 128 else "." for i in range(256)}


def hexdump(data: bytes, linesize: int) -> list[str]:
    """
    data is a bytes object. The returned result is a string.
    """
    prettylines = []
    if len(data) < 65536:
        fmt = "%%04X   %%-%ds   %%s"
    else:
        fmt = "%%08X   %%-%ds   %%s"
    fmt = fmt % (3 * linesize - 1,)
    for i in range(0, len(data), linesize):
        line = data[i : i + linesize]
        hextext = line.hex(" ")
        rawtext = "".join(_printable[b] for b in line)
        prettylines.append(fmt % (i, hextext, rawtext))
    return prettylines


class HexString(bytes):
    """
    Represents bytes that will be hex-dumped to a string when its string
    representation is requested.
    """
    def __init__(self, data: bytes, linesize: int = 16) -> None:
        self.linesize = linesize

    def __new__(cls, data: bytes, *args: object, **kwargs: object) -> Self:
        return bytes.__new__(cls, data)

    def __str__(self) -> str:
        if not self:
            return "''"
        sep = "\n"
        return sep + sep.join(
            hexdump(self, self.linesize))



#-------------------------------------------------------------------------------
# elftools: dwarf/typeunit.py
#
# DWARF type unit
#
# Dinkar Khandalekar (contact@dinkar.dev)
# This code is in the public domain
#-------------------------------------------------------------------------------
from __future__ import annotations

from bisect import bisect_right
from functools import cached_property
from typing import TYPE_CHECKING, Any

from .die import DIE
from ..common.utils import dwarf_assert

if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..construct.lib.container import Container
    from .abbrevtable import AbbrevTable
    from .dwarfinfo import DWARFInfo
    from .structs import DWARFStructs


class TypeUnit:
    """ A DWARF type unit (TU).

            A type unit contains type definition entries that can be used to
            reference to type definition for debugging information entries in
            other compilation units and type units. Each type unit must be uniquely
            identified by a 64-bit signature. (DWARFv4 section 3.1.3)

            Type units are stored in the .debug_types section. This section was
            introduced by the DWARFv4 standard (and removed in the DWARFv5 standard;
            the underlying type units were relocated to the .debug_info
            section - DWARFv5 section 1.4)

        Serves as a container and context to DIEs that describe type definitions
        referenced from compilation units and other type units.

        TU header entries can be accessed as dict keys from this object, i.e.
           tu = TypeUnit(...)
           tu['version']  # version field of the TU header

        To get the top-level DIE describing the type unit, call the
        get_top_DIE method.
    """
    def __init__(
        self,
        header: Container,
        dwarfinfo: DWARFInfo,
        structs: DWARFStructs,
        tu_offset: int,
        tu_die_offset: int,
    ) -> None:
        """ header:
                TU header for this type unit

            dwarfinfo:
                The DWARFInfo context object which created this one

            structs:
                A DWARFStructs instance suitable for this type unit

            tu_offset:
                Offset in the stream to the beginning of this TU (its header)

            tu_die_offset:
                Offset in the stream of the top DIE of this TU
        """
        self.dwarfinfo = dwarfinfo
        self.header = header
        self.structs = structs
        self.tu_offset = tu_offset
        self.tu_die_offset = tu_die_offset

        # A list of DIEs belonging to this TU.
        # This list is lazily constructed as DIEs are iterated over.
        self._dielist: list[DIE] = []
        # A list of file offsets, corresponding (by index) to the DIEs
        # in `self._dielist`. This list exists separately from
        # `self._dielist` to make it binary searchable, enabling the
        # DIE population strategy used in `iter_DIE_children`.
        # Like `self._dielist`, this list is lazily constructed
        # as DIEs are iterated over.
        self._diemap: list[int] = []

    @property
    def cu_offset(self) -> int:
        """Simulates the cu_offset attribute required by the DIE by returning the tu_offset instead
        """
        return self.tu_offset

    @property
    def cu_die_offset(self) -> int:
        """Simulates the cu_die_offset attribute required by the DIE by returning the tu_offset instead
        """
        return self.tu_die_offset

    def dwarf_format(self) -> int:
        """ Get the DWARF format (32 or 64) for this TU
        """
        return self.structs.dwarf_format

    def get_abbrev_table(self) -> AbbrevTable:
        """ Get the abbreviation table (AbbrevTable object) for this TU
        """
        return self._abbrev_table

    @cached_property
    def _abbrev_table(self) -> AbbrevTable:
        return self.dwarfinfo.get_abbrev_table(self['debug_abbrev_offset'])

    def get_top_DIE(self) -> DIE:
        """ Get the top DIE (which is DW_TAG_type_unit entry) of this TU
        """

        # Note that a top DIE always has minimal offset and is therefore
        # at the beginning of our lists, so no bisect is required.
        if self._diemap:
            return self._dielist[0]

        assert self.dwarfinfo.debug_types_sec is not None
        top = DIE(
                cu=self,
                stream=self.dwarfinfo.debug_types_sec.stream,
                offset=self.tu_die_offset)

        self._dielist.insert(0, top)
        self._diemap.insert(0, self.tu_die_offset)

        top._translate_indirect_attributes()  # Can't translate indirect attributes until the top DIE has been parsed to the end

        return top

    def has_top_DIE(self) -> bool:
        """ Returns whether the top DIE in this TU has already been parsed and cached.
            No parsing on demand!
        """
        return bool(self._diemap)

    @property
    def size(self) -> int:
        return self['unit_length'] + self.structs.initial_length_field_size()

    def iter_DIEs(self) -> Iterator[DIE]:
        """ Iterate over all the DIEs in the TU, in order of their appearance.
            Note that null DIEs will also be returned.
        """
        return self._iter_DIE_subtree(self.get_top_DIE())

    def iter_DIE_children(self, die: DIE) -> Iterator[DIE]:
        """ Given a DIE, yields either its children, without null DIE list
            terminator, or nothing, if that DIE has no children.

            The null DIE terminator is saved in that DIE when iteration ended.
        """
        if not die.has_children:
            return

        # `cur_offset` tracks the stream offset of the next DIE to yield
        # as we iterate over our children,
        cur_offset = die.offset + die.size

        while True:
            child = self._get_cached_DIE(cur_offset)

            child.set_parent(die)

            if child.is_null():
                die._terminator = child
                return

            yield child

            if not child.has_children:
                cur_offset += child.size
            elif "DW_AT_sibling" in child.attributes:
                sibling = child.attributes["DW_AT_sibling"]
                if sibling.form in ('DW_FORM_ref1', 'DW_FORM_ref2',
                                    'DW_FORM_ref4', 'DW_FORM_ref8',
                                    'DW_FORM_ref', 'DW_FORM_ref_udata'):
                    cur_offset = sibling.value + self.tu_offset
                elif sibling.form == 'DW_FORM_ref_addr':
                    cur_offset = sibling.value
                else:
                    raise NotImplementedError('sibling in form %s' % sibling.form)
            else:
                # If no DW_AT_sibling attribute is provided by the producer
                # then the whole child subtree must be parsed to find its next
                # sibling. There is one zero byte representing null DIE
                # terminating children list. It is used to locate child subtree
                # bounds.

                # If children are not parsed yet, this instruction will manage
                # to recursive call of this function which will result in
                # setting of `_terminator` attribute of the `child`.
                if child._terminator is None:
                    for _ in self.iter_DIE_children(child):
                        pass
                    assert child._terminator is not None

                cur_offset = child._terminator.offset + child._terminator.size

    def get_DIE_from_refaddr(self, refaddr: int) -> DIE:
        """ Obtain a DIE contained in this CU from a reference.
            refaddr:
                The offset into the .debug_info section, which must be
                contained in this CU or a DWARFError will be raised.
            When using a reference class attribute with a form that is
            relative to the compile unit, add unit add the compile unit's
            .cu_addr before calling this function.
        """
        # All DIEs are after the cu header and within the unit
        dwarf_assert(
            self.cu_die_offset <= refaddr < self.cu_offset + self.size,
            'refaddr %s not in DIE range of CU %s' % (refaddr, self.cu_offset))

        return self._get_cached_DIE(refaddr)

    #------ PRIVATE ------#

    def __getitem__(self, name: str) -> Any:
        """ Implement dict-like access to header entries
        """
        return self.header[name]

    def _iter_DIE_subtree(self, die: DIE) -> Iterator[DIE]:
        """ Given a DIE, this yields it with its subtree including null DIEs
            (child list terminators).
        """
        # If the die is an imported unit, replace it with what it refers to if
        # we can
        if die.tag == 'DW_TAG_imported_unit' and self.dwarfinfo.supplementary_dwarfinfo:
            die = die.get_DIE_from_attribute('DW_AT_import')
        yield die
        if die.has_children:
            for c in die.iter_children():
                yield from die.cu._iter_DIE_subtree(c)
            assert die._terminator is not None
            yield die._terminator

    def _get_cached_DIE(self, offset: int) -> DIE:
        """ Given a DIE offset, look it up in the cache.  If not present,
            parse the DIE and insert it into the cache.

            offset:
                The offset of the DIE in the debug_types section to retrieve.

            The stream reference is copied from the top DIE.  The top die will
            also be parsed and cached if needed.

            See also get_DIE_from_refaddr(self, refaddr).
        """
        # The top die must be in the cache if any DIE is in the cache.
        # The stream is the same for all DIEs in this TU, so populate
        # the top DIE and obtain a reference to its stream.
        top_die_stream = self.get_top_DIE().stream

        # `offset` is the offset in the stream of the DIE we want to return.
        # The map is maintined as a parallel array to the list.  We call
        # bisect each time to ensure new DIEs are inserted in the correct
        # order within both `self._dielist` and `self._diemap`.
        i = bisect_right(self._diemap, offset)

        # Note that `self._diemap` cannot be empty because a the top DIE
        # was inserted by the call to .get_top_DIE().  Also it has the minimal
        # offset, so the bisect_right insert point will always be at least 1.
        if offset == self._diemap[i - 1]:
            die = self._dielist[i - 1]
        else:
            die = DIE(cu=self, stream=top_die_stream, offset=offset)
            self._dielist.insert(i, die)
            self._diemap.insert(i, offset)

        return die



#-------------------------------------------------------------------------------
# elftools: elf/hash.py
#
# ELF hash table sections
#
# Andreas Ziegler (andreas.ziegler@fau.de)
# This code is in the public domain
#-------------------------------------------------------------------------------
from __future__ import annotations

import struct
from typing import TYPE_CHECKING, Protocol

from ..common.utils import struct_parse
from ..construct.lib.container import Container
from .sections import Section

if TYPE_CHECKING:
    from .elffile import ELFFile
    from .sections import Symbol


class _SymbolTable(Protocol):
    """Common base-class of elftools.elf.sections.SymbolTableSection and
    elftools.elf.dynamic.DynamicSegment to be consumed by
    (ELF|GNU)Hash(Section|Table)."""
    def get_symbol(self, index: int, /) -> Symbol | None: ...


class ELFHashTable:
    """ Representation of an ELF hash table to find symbols in the
        symbol table - useful for super-stripped binaries without section
        headers where only the start of the symbol table is known from the
        dynamic segment. The layout and contents are nicely described at
        https://flapenguin.me/2017/04/24/elf-lookup-dt-hash/.

        The symboltable argument needs to implement a get_symbol() method -
        in a regular ELF file, this will be the linked symbol table section
        as indicated by the sh_link attribute. For super-stripped binaries,
        one should use the DynamicSegment object as the symboltable as it
        supports symbol lookup without access to a symbol table section.
    """

    def __init__(
        self,
        elffile: ELFFile,
        start_offset: int,
        size: int | None,
        symboltable: _SymbolTable,
    ) -> None:
        """
        Args:
            elffile (ELFFile): The ELF file.
            start_offset: The offset of the start of the symbol table in the ELF file.
            size: Size of the table in bytes (can be None if unknown).
            symboltable: A symbol table with a get_symbol() method to do symbol lookup.
        """
        self.elffile = elffile
        self._symboltable = symboltable
        if size == 0:  # size may also be None if its unknown
            self.params = Container(**{
                'nbuckets': 0,
                'nchains': 0,
                'buckets': [],
                'chains': [],
            })
        else:
            self.params = struct_parse(self.elffile.structs.Elf_Hash,
                                       self.elffile.stream,
                                       start_offset)

    def get_number_of_symbols(self) -> int:
        """ Get the number of symbols from the hash table parameters.
        """
        return self.params['nchains']

    def get_symbol(self, name: str) -> Symbol | None:
        """ Look up a symbol from this hash table with the given name.
        """
        if self.params['nbuckets'] == 0:
            return None
        hval = self.elf_hash(name) % self.params['nbuckets']
        symndx = self.params['buckets'][hval]
        while symndx != 0:
            sym = self._symboltable.get_symbol(symndx)
            if sym and sym.name == name:
                return sym
            symndx = self.params['chains'][symndx]
        return None

    @staticmethod
    def elf_hash(name: bytes | str) -> int:
        """ Compute the hash value for a given symbol name.
        """
        if not isinstance(name, bytes):
            name = name.encode('utf-8')
        h = 0
        x = 0
        for c in bytearray(name):
            h = (h << 4) + c
            x = h & 0xF0000000
            if x != 0:
                h ^= (x >> 24)
            h &= ~x
        return h


class ELFHashSection(Section, ELFHashTable):
    """ Section representation of an ELF hash table. In regular ELF files, this
        allows us to use the common functions defined on Section objects when
        dealing with the hash table.
    """
    def __init__(
        self,
        header: Container,
        name: str,
        elffile: ELFFile,
        symboltable: _SymbolTable,
    ) -> None:
        Section.__init__(self, header, name, elffile)
        ELFHashTable.__init__(self, elffile, self['sh_offset'], self['sh_size'], symboltable)


class GNUHashTable:
    """ Representation of a GNU hash table to find symbols in the
        symbol table - useful for super-stripped binaries without section
        headers where only the start of the symbol table is known from the
        dynamic segment. The layout and contents are nicely described at
        https://flapenguin.me/2017/05/10/elf-lookup-dt-gnu-hash/.

        The symboltable argument needs to implement a get_symbol() method -
        in a regular ELF file, this will be the linked symbol table section
        as indicated by the sh_link attribute. For super-stripped binaries,
        one should use the DynamicSegment object as the symboltable as it
        supports symbol lookup without access to a symbol table section.
    """
    def __init__(
        self,
        elffile: ELFFile,
        start_offset: int,
        symboltable: _SymbolTable,
    ) -> None:
        self.elffile = elffile
        self._symboltable = symboltable
        self.params: Container = struct_parse(self.elffile.structs.Gnu_Hash,
                                   self.elffile.stream,
                                   start_offset)
        # Element sizes in the hash table
        self._wordsize: int = self.elffile.structs.Elf_word('').sizeof()
        self._xwordsize: int = self.elffile.structs.Elf_xword('').sizeof()
        self._chain_pos: int = start_offset + 4 * self._wordsize + \
            self.params['bloom_size'] * self._xwordsize + \
            self.params['nbuckets'] * self._wordsize

    def get_number_of_symbols(self) -> int:
        """ Get the number of symbols in the hash table by finding the bucket
            with the highest symbol index and walking to the end of its chain.
        """
        # Find highest index in buckets array
        max_idx = max(self.params['buckets'])
        if max_idx < self.params['symoffset']:
            return self.params['symoffset']

        # Position the stream at the start of the corresponding chain
        max_chain_pos = self._chain_pos + \
            (max_idx - self.params['symoffset']) * self._wordsize
        self.elffile.stream.seek(max_chain_pos)
        hash_format = '<I' if self.elffile.little_endian else '>I'

        # Walk the chain to its end (lowest bit is set)
        while True:
            cur_hash = struct.unpack(hash_format, self.elffile.stream.read(self._wordsize))[0]
            if cur_hash & 1:
                return max_idx + 1

            max_idx += 1

    def _matches_bloom(self, H1: int) -> bool:
        """ Helper function to check if the given hash could be in the hash
            table by testing it against the bloom filter.
        """
        arch_bits = self.elffile.elfclass
        H2 = H1 >> self.params['bloom_shift']
        word_idx = int(H1 / arch_bits) % self.params['bloom_size']
        BITMASK = (1 << (H1 % arch_bits)) | (1 << (H2 % arch_bits))
        return (self.params['bloom'][word_idx] & BITMASK) == BITMASK

    def get_symbol(self, name: str) -> Symbol | None:
        """ Look up a symbol from this hash table with the given name.
        """
        namehash = self.gnu_hash(name)
        if not self._matches_bloom(namehash):
            return None

        symidx = self.params['buckets'][namehash % self.params['nbuckets']]
        if symidx < self.params['symoffset']:
            return None

        self.elffile.stream.seek(self._chain_pos + (symidx - self.params['symoffset']) * self._wordsize)
        hash_format = '<I' if self.elffile.little_endian else '>I'
        while True:
            cur_hash = struct.unpack(hash_format, self.elffile.stream.read(self._wordsize))[0]
            if cur_hash | 1 == namehash | 1:
                symbol = self._symboltable.get_symbol(symidx)
                if symbol and name == symbol.name:
                    return symbol

            if cur_hash & 1:
                break
            symidx += 1
        return None

    @staticmethod
    def gnu_hash(key: bytes | str) -> int:
        """ Compute the GNU-style hash value for a given symbol name.
        """
        if not isinstance(key, bytes):
            key = key.encode('utf-8')
        h = 5381
        for c in bytearray(key):
            h = h * 33 + c
        return h & 0xFFFFFFFF


class GNUHashSection(Section, GNUHashTable):
    """ Section representation of a GNU hash table. In regular ELF files, this
        allows us to use the common functions defined on Section objects when
        dealing with the hash table.
    """
    def __init__(
        self,
        header: Container,
        name: str,
        elffile: ELFFile,
        symboltable: _SymbolTable,
    ) -> None:
        Section.__init__(self, header, name, elffile)
        GNUHashTable.__init__(self, elffile, self['sh_offset'], symboltable)



#-------------------------------------------------------------------------------
# elftools: common/construct_utils.py
#
# Some complementary construct utilities
#
# Eli Bendersky (eliben@gmail.com)
# This code is in the public domain
#-------------------------------------------------------------------------------
from __future__ import annotations

from struct import Struct
from typing import IO, TYPE_CHECKING, Any, NoReturn

from ..construct import (
    Subconstruct, ConstructError, ArrayError, SizeofError, Construct, StaticField, FieldError
    )

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable

    from ..construct import Container


class RepeatUntilExcluding(Subconstruct):
    """ A version of construct's RepeatUntil that doesn't include the last
        element (which casued the repeat to exit) in the return value.

        Only parsing is currently implemented.

        P.S. removed some code duplication
    """
    __slots__ = ("predicate",)
    def __init__(self, predicate: Callable[[Any, Container], bool], subcon: Construct) -> None:
        Subconstruct.__init__(self, subcon)
        self.predicate = predicate
        self._clear_flag(self.FLAG_COPY_CONTEXT)
        self._set_flag(self.FLAG_DYNAMIC)
    def _parse(self, stream: IO[bytes], context: Container) -> list[Any]:
        obj = []
        try:
            context_for_subcon = context
            if self.subcon.conflags & self.FLAG_COPY_CONTEXT:
                context_for_subcon = context.__copy__()

            while True:
                subobj = self.subcon._parse(stream, context_for_subcon)
                if self.predicate(subobj, context):
                    break
                obj.append(subobj)
        except ConstructError as ex:
            raise ArrayError("missing terminator", ex)
        return obj
    def _build(self, obj: Iterable[Any], stream: IO[bytes], context: Container) -> NoReturn:
        raise NotImplementedError('no building')
    def _sizeof(self, context: Container) -> int:
        raise SizeofError("can't calculate size")


class _NamedConstruct(Construct):
    if TYPE_CHECKING:
        name: str  # instead of `str|None` from Construct to save us from `is None` checks everywhere


class ULEB128(_NamedConstruct):
    """A construct based parser for ULEB128 encoding.
    """
    def _parse(self, stream: IO[bytes], context: Container) -> int:
        value = 0
        shift = 0
        while True:
            data = stream.read(1)
            if len(data) != 1:
                raise FieldError("unexpected end of stream while parsing a ULEB128 encoded value")
            b = data[0]
            value |= (b & 0x7F) << shift
            shift += 7
            if b & 0x80 == 0:
                return value


class SLEB128(_NamedConstruct):
    """A construct based parser for SLEB128 encoding.
    """
    def _parse(self, stream: IO[bytes], context: Container) -> int:
        value = 0
        shift = 0
        while True:
            data = stream.read(1)
            if len(data) != 1:
                raise FieldError("unexpected end of stream while parsing a SLEB128 encoded value")
            b = data[0]
            value |= (b & 0x7F) << shift
            shift += 7
            if b & 0x80 == 0:
                return value | (~0 << shift) if b & 0x40 else value


class StreamOffset(_NamedConstruct):
    """
    Captures the current stream offset

    Parameters:
    * name - the name of the value

    Example:
    StreamOffset("item_offset")
    """
    __slots__ = ()
    def __init__(self, name: str) -> None:
        Construct.__init__(self, name)
        self._set_flag(self.FLAG_DYNAMIC)
    def _parse(self, stream: IO[bytes], context: Container) -> int:
        return stream.tell()
    def _build(self, obj: None, stream: IO[bytes], context: Container) -> None:
        context[self.name] = stream.tell()
    def _sizeof(self, context: Container) -> int:
        return 0

_UBInt24_packer = Struct(">BH")
_ULInt24_packer = Struct("<HB")

class UBInt24(StaticField):
    """unsigned, big endian 24-bit integer"""
    def __init__(self, name: str) -> None:
        StaticField.__init__(self, name, 3)

    def _parse(self, stream: IO[bytes], context: Container) -> int:
        (h, l) = _UBInt24_packer.unpack(StaticField._parse(self, stream, context))
        return l | (h << 16)

    def _build(self, obj: int, stream: IO[bytes], context: Container) -> None:
        StaticField._build(self, _UBInt24_packer.pack(obj >> 16, obj & 0xFFFF), stream, context)

class ULInt24(StaticField):
    """unsigned, little endian 24-bit integer"""
    def __init__(self, name: str) -> None:
        StaticField.__init__(self, name, 3)

    def _parse(self, stream: IO[bytes], context: Container) -> int:
        (l, h) = _ULInt24_packer.unpack(StaticField._parse(self, stream, context))
        return l | (h << 16)

    def _build(self, obj: int, stream: IO[bytes], context: Container) -> None:
        StaticField._build(self, _ULInt24_packer.pack(obj & 0xFFFF, obj >> 16), stream, context)



#-------------------------------------------------------------------------------
# elftools: common/utils.py
#
# Miscellaneous utilities for elftools
#
# Eli Bendersky (eliben@gmail.com)
# This code is in the public domain
#-------------------------------------------------------------------------------
from __future__ import annotations

from contextlib import contextmanager
from typing import IO, TYPE_CHECKING, Any, TypeVar, overload

from .exceptions import ELFParseError, ELFError, DWARFError
from ..construct import ConstructError
import os

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping

    from ..construct import Construct, FormatField
    from ..dwarf.dwarfinfo import DebugSectionDescriptor
    from .construct_utils import SLEB128, ULEB128, UBInt24, ULInt24

    _T = TypeVar("_T")
    _K = TypeVar("_K")
    _V = TypeVar("_V")


def merge_dicts(*dicts: Mapping[_K, _V]) -> dict[_K, _V]:
    "Given any number of dicts, merges them into a new one."""
    result: dict[_K, _V] = {}
    for d in dicts:
        result.update(d)
    return result

def bytes2str(b: bytes) -> str:
    """Decode a bytes object into a string."""
    return b.decode('latin-1')


# Use @overload to get more specific type, e.g. [SU][BLN]{EB,Int}{8,16,24,32,64,128} -> int
@overload
def struct_parse(struct: FormatField[_T] | ULEB128 | SLEB128 | UBInt24 | ULInt24, stream: IO[bytes], stream_pos: int | None = ...) -> _T: ...
@overload
def struct_parse(struct: Construct, stream: IO[bytes], stream_pos: int | None = ...) -> Any: ...
def struct_parse(struct: Construct, stream: IO[bytes], stream_pos: int | None = None) -> Any:
    """ Convenience function for using the given struct to parse a stream.
        If stream_pos is provided, the stream is seeked to this position before
        the parsing is done. Otherwise, the current position of the stream is
        used.
        Wraps the error thrown by construct with ELFParseError.
    """
    try:
        if stream_pos is not None:
            stream.seek(stream_pos)
        return struct.parse_stream(stream)
    except ConstructError as e:
        raise ELFParseError(str(e))


def parse_cstring_from_stream(stream: IO[bytes], stream_pos: int | None = None) -> bytes | None:
    """ Parse a C-string from the given stream. The string is returned without
        the terminating \x00 byte. If the terminating byte wasn't found, None
        is returned (the stream is exhausted).
        If stream_pos is provided, the stream is seeked to this position before
        the parsing is done. Otherwise, the current position of the stream is
        used.
        Note: a bytes object is returned here, because this is what's read from
        the binary file.
    """
    if stream_pos is not None:
        stream.seek(stream_pos)
    CHUNKSIZE = 64
    chunks = []
    while True:
        chunk, sep, _tail = stream.read(CHUNKSIZE).partition(b'\x00')
        chunks.append(chunk)
        if sep:
            return b''.join(chunks)
        if len(chunk) < CHUNKSIZE:
            return None


def elf_assert(cond: object, msg: str = '') -> None:
    """ Assert that cond is True, otherwise raise ELFError(msg)
    """
    _assert_with_exception(cond, msg, ELFError)


def dwarf_assert(cond: object, msg: str = '') -> None:
    """ Assert that cond is True, otherwise raise DWARFError(msg)
    """
    _assert_with_exception(cond, msg, DWARFError)


@contextmanager
def preserve_stream_pos(stream: IO[bytes]) -> Iterator[None]:
    """ Usage:
        # stream has some position FOO (return value of stream.tell())
        with preserve_stream_pos(stream):
            # do stuff that manipulates the stream
        # stream still has position FOO
    """
    saved_pos = stream.tell()
    yield
    stream.seek(saved_pos)


def roundup(num: int, bits: int) -> int:
    """ Round up a number to nearest multiple of 2^bits. The result is a number
        where the least significant bits passed in bits are 0.
    """
    return (num - 1 | (1 << bits) - 1) + 1


def save_dwarf_section(section: DebugSectionDescriptor, filename: str) -> None:
    """Debug helper: dump section contents into a file
    Section is expected to be one of the debug_xxx_sec elements of DWARFInfo
    """
    stream = section.stream
    with preserve_stream_pos(stream), open(filename, 'wb') as file:
        stream.seek(0, os.SEEK_SET)
        data = stream.read(section.size)
        file.write(data)


#------------------------- PRIVATE -------------------------

def _assert_with_exception(cond: object, msg: str, exception_type: type[BaseException]) -> None:
    if not cond:
        raise exception_type(msg)



#-------------------------------------------------------------------------------
# elftools: dwarf/abbrevtable.py
#
# DWARF abbreviation table
#
# Eli Bendersky (eliben@gmail.com)
# This code is in the public domain
#-------------------------------------------------------------------------------
from __future__ import annotations

from typing import IO, TYPE_CHECKING, Any

from ..common.utils import struct_parse

if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..construct.lib.container import Container
    from .structs import DWARFStructs


class AbbrevTable:
    """ Represents a DWARF abbreviation table.
    """
    __slots__ = ('structs', 'stream', 'offset', '_abbrev_map')
    def __init__(self, structs: DWARFStructs, stream: IO[bytes], offset: int) -> None:
        """ Create new abbreviation table. Parses the actual table from the
            stream and stores it internally.

            structs:
                A DWARFStructs instance for parsing the data

            stream, offset:
                The stream and offset into the stream where this abbreviation
                table lives.
        """
        self.structs = structs
        self.stream = stream
        self.offset = offset

        self._abbrev_map = self._parse_abbrev_table()

    def get_abbrev(self, code: int) -> AbbrevDecl:
        """ Get the AbbrevDecl for a given code. Raise KeyError if no
            declaration for this code exists.
        """
        return self._abbrev_map[code]

    def _parse_abbrev_table(self) -> dict[int, AbbrevDecl]:
        """ Parse the abbrev table from the stream
        """
        map: dict[int, AbbrevDecl] = {}
        self.stream.seek(self.offset)
        while True:
            decl_code: int = struct_parse(
                struct=self.structs.the_Dwarf_uleb128,
                stream=self.stream)
            if decl_code == 0:
                break
            declaration = struct_parse(
                struct=self.structs.Dwarf_abbrev_declaration,
                stream=self.stream)
            map[decl_code] = AbbrevDecl(decl_code, declaration)
        return map


class AbbrevDecl:
    """ Wraps a parsed abbreviation declaration, exposing its fields with
        dict-like access, and adding some convenience methods.

        The abbreviation declaration represents an "entry" that points to it.
    """
    __slots__ = ('code', 'decl', '_has_children')
    def __init__(self, code: int, decl: Container) -> None:
        self.code = code
        self.decl = decl
        self._has_children = decl['children_flag'] == 'DW_CHILDREN_yes'

    def has_children(self) -> bool:
        return self._has_children

    def iter_attr_specs(self) -> Iterator[tuple[str, str]]:
        """ Iterate over the attribute specifications for the entry. Yield
            (name, form) pairs.
        """
        for attr_spec in self['attr_spec']:
            yield attr_spec.name, attr_spec.form

    def __getitem__(self, entry: str) -> Any:
        return self.decl[entry]
